package qa.qcri.prepocressor.operations;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;

import qa.qcri.prepocressor.datastructures.Individual;
import qa.qcri.prepocressor.imageprocessing.Geometry;
import qa.qcri.prepocressor.ui.Configuration;

/*
This file is part of the prepocressor toolkit.

Copyright (c) 2015, QCRI a member of Qatar Foundation. All Rights Reserved
Felix Stahlberg

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 * This class implements the vertTextSegmentation operation.
 * 
 * @author Felix Stahlberg
 */
public class VertTextSegmentationOperation extends Operation {

	/* (non-Javadoc)
	 * @see Operation#Operation(String)
	 */
	public VertTextSegmentationOperation(String name) {
		super(name);
	}

	/* (non-Javadoc)
	 * @see Operation#createConfiguration()
	 */
	@Override
	protected Configuration createConfiguration() {
		return new Configuration() {
			@Override
			protected void addDefaultParameters() {
				addParameter("minSlope", 0.25f,
					"Minimal gradient in the vertical projection of a text "
					+ "area. The values of the vertical projection range "
					+ "from 0 to 1, and the projection is resized to "
					+ "-resolution");
				addParameter("minWidth", 0.1f,
					"Minimal width of a text area relative to the page width.");
				addParameter("dilate", 0.05f,
					"Safety margin which is added to the found text segments, "
					+ "relative to the page width. Set to negative to disable "
					+ "dilation.");
				addParameter("outputSegmentation", 0,
					"Output an 1xresolution array indicating the classification"
					+ " of the columns in text and non-text columns. If this "
					+ "is set to 0, the input image is cutted according the "
					+ "text segmentations.");
				addParameter("minMargin", 0.05f,
					"Minimal vertical distance between two text areas "
					+ "relative to the page width.");
				addParameter("resolution", 100,
					"Resolution for the vertical projection. 100 is a good "
					+ "value even for documents with largely differing sizes.");
				addParameter("morph", "openFirst",
					"Morphology operations on the segmentation.\n+\n"
					+ "* openFirst: opening, then closing\n"
					+ "* closeFirst: closing, then opening\n"
					+ "* none: No morphology operation.\n-");
				addParameter("concatChildren", 0,
					"Set to 0 to pass each found text segment separately "
					+ "through the pipeline. Set to 1 to concat all children "
					+ "to a single image removing the non-text areas.");
			}
			@Override
			protected String getDescription() {
				return "Extracts vertical cuts of text areas. The areas are "
					+ "identified by a constant horizontal gradient of the "
					+ "vertical projection.";
			}
		};
	}

	@Override
	protected List<Individual> processIndividual(Individual indy) {
		Mat content = indy.getContent();
		int res = conf.getInteger("resolution");
		Mat threshold = getThresholdImage(content);
		// Cut image
		List<Individual> children = new LinkedList<Individual>();
		if (conf.getInteger("outputSegmentation") > 0) {
			Individual child = new Individual();
			child.setContent(threshold);
			children.add(child);
			return children;
		}
		for (int from = 0; from < res; from++) {
			if (threshold.get(0, from)[0] > 0.5) {
				int to = from+1;
				for (; to < res; to++) {
					if (threshold.get(0, to)[0] < 0.5) {
						break;
					}
				}
				children.add(indy.produceChild(new Rect(
					new Point(((double) from)/((double) res)*content.cols(), 0),
					new Point(((double) to)/((double) res)*content.cols(), 
							content.rows())
				)));
				from = to;
			}
		}
		if (conf.getInteger("concatChildren") > 0 && children.size() > 1) {
			Individual child = new Individual();
			child.setContent(concatChildren(indy, children));
			children.clear();
			children.add(child);
		} else {
			Collections.reverse(children); // Because of arabic right-to-left
		}
		// TODO: Release content?
		threshold.release();
		return children;
	}
	
	/* (non-Javadoc)
	 * Combines children into a single image
	 */
	private Mat concatChildren(Individual indy, List<Individual> children) {
		Mat content = indy.getContent();
		int width = 0;
		Point p1 = new Point(1000000, 1000000);
		Point p2 = new Point(-1000000, -1000000);
		for (Individual child : children) {
			width += child.getContent().cols();
			Rect childRect = child.getParentRect();
			p1.x = Math.min(p1.x, childRect.x);
			p1.y = Math.min(p1.y, childRect.y);
			p2.x = Math.max(p2.x, childRect.x + childRect.width);
			p2.y = Math.max(p2.y, childRect.y + childRect.height);
		}
		Mat dst = Mat.zeros(content.rows(), width, content.type());
		int border = 0;
		for (Individual childIndy : children) {
			Mat child = childIndy.getContent();
			Mat tmp = new Mat(child.rows(), child.cols(), child.type());
			child.copyTo(tmp);
			Mat tmpBorder = new Mat(content.rows(), width, content.type());
			Imgproc.copyMakeBorder(tmp, tmpBorder, 0, 0,
					border, width - border - child.cols(),
					Imgproc.BORDER_CONSTANT);
			Core.add(dst, tmpBorder, dst);
			border += child.cols();
		}
		Individual child = new Individual();
		child.setContent(dst);
		child.setParent(indy, new Rect(p1, p2));
		return dst;
	}
	
	/* (non-Javadoc)
	 * get threshold image based on projection profile derivative
	 */
	private Mat getThresholdImage(Mat content) {
		Mat red = new Mat(content.rows(), content.cols(), CvType.CV_32FC1);
		Mat threshold = new Mat(content.rows(), content.cols(), CvType.CV_8UC1);
		Mat accSobel = new Mat(content.rows(), content.cols(), CvType.CV_32FC1);
		Mat accNorm = new Mat(content.rows(), content.cols(), CvType.CV_32FC1);
		Mat acc100 = new Mat(content.rows(), content.cols(), CvType.CV_32FC1);
		int res = conf.getInteger("resolution");

		Core.reduce(content, red, 0, Core.REDUCE_SUM, CvType.CV_32FC1);
		Imgproc.resize(Geometry.rectSum(red, 0, 0, 1, red.cols()),
			acc100, new Size(res, 1), 0, 0, Imgproc.INTER_LINEAR);
		Core.normalize(acc100, accNorm, 0.0, 1.0, Core.NORM_MINMAX);
		Imgproc.Sobel(accNorm, accSobel, -1,
				1, 0, 3, 1, 0, Imgproc.BORDER_REFLECT);
		Imgproc.threshold(accSobel, threshold, conf.getFloat("minSlope"), 255,
				Imgproc.THRESH_BINARY);
		red.release();
		acc100.release();
		accNorm.release();
		accSobel.release();
		if (!conf.getString("morph").equals("none")) {
			int closeSize = (int) (conf.getFloat("minMargin")*res);
			int openSize = (int) (conf.getFloat("minWidth")*res);
			Mat closingKernel = Imgproc.getStructuringElement(Imgproc.MORPH_RECT,
					new Size(Math.max(1, closeSize), 1),
					new Point(closeSize/2, 0));
			Mat openingKernel = Imgproc.getStructuringElement(Imgproc.MORPH_RECT,
					new Size(Math.max(1, openSize), 1),
					new Point(openSize/2, 0));
			if (conf.getString("morph").equals("closeFirst")) {
				Imgproc.morphologyEx(threshold, threshold, Imgproc.MORPH_CLOSE,
						closingKernel);
				Imgproc.morphologyEx(threshold, threshold, Imgproc.MORPH_OPEN,
						openingKernel);
			} else {
				Imgproc.morphologyEx(threshold, threshold, Imgproc.MORPH_OPEN,
						openingKernel);
				Imgproc.morphologyEx(threshold, threshold, Imgproc.MORPH_CLOSE,
						closingKernel);
			}
		}
		int dilateSize = (int) (res*conf.getFloat("dilate"));
		if (dilateSize > 0) {
			Mat dilateKernel = Imgproc.getStructuringElement(Imgproc.MORPH_RECT,
					new Size(dilateSize, 1),
					new Point(dilateSize/2, 0));
			Imgproc.morphologyEx(threshold, threshold, Imgproc.MORPH_DILATE,
				dilateKernel);
		}
		return threshold;
	}
}
